package Stanford::LDAP::EntriesDB;

use DBI;
use Data::Dumper;
use Carp;

# Make this a Moose class
use Moose ;
use namespace::autoclean;

has 'db_file' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default => '/tmp/sugal_sync.sqlite',
) ;

has 'dbh' => (
  is        => 'rw',
  isa       => 'Any',
  required  => 0,
) ;

has 'table_name' => (
  is        => 'rw',
  isa       => 'Str',
    required  => 1,
    default => 'elements',
) ;

# Width of the dn field
has 'dn_width' => (
  is        => 'rw',
  isa       => 'Int',
    required  => 1,
    default => '255',
) ;



# Create a connection to the entriesDB
my $entriesDB_file = '/tmp/sugal_sync.sqlite';
sub connect {
    my $self = shift();

    my $dbh = DBI->connect("dbi:SQLite:dbname=$entriesDB_file", q{}, q{});
    $self->dbh($dbh);

    return $self->dbh($dbh);
}

sub create_db {
    my $self = shift();

    # Delete the database file.
    if ($self->db_file() && (-e $self->db_file())) {
        unlink($self->db_file());
    }

    # Create the database.
    my $dbh = $self->connect();

    my $table_name = $self->table_name();
    my $dn_width   = $self->dn_width();

    # Create the table.
    my $qry = <<"EOT";
CREATE TABLE $table_name(
    id INTEGER PRIMARY KEY ASC,
    dn CHAR($dn_width)
    )
EOT

    my $sth = $dbh->prepare($qry);
    $sth->execute();

    return;
}

# Read N entries from the database.
sub read_entries {
    my $self = shift();
    my ($N) = @_;

    if (!defined($N) || ($N == 0)) {
        croak "missing limit";
    }

    my $table_name = $self->table_name();

    my $qry = <<"EOQ";
    SELECT id, dn
        FROM $table_name
        LIMIT $N
EOQ

    my $sth = $self->dbh()->prepare($qry);
    $sth->execute();

    my @results = ();
    while (my @data = $sth->fetchrow_array()) {
        my $id = $data[0];
        my $dn = $data[1];
        push(@results, [$id, $dn]);
    }

    return @results;
}

sub delete_entries {
    my $self = shift();
    my ($entries_to_delete_aref) = @_;

    my @entries_to_delete = @{ $entries_to_delete_aref };

    my $table_name = $self->table_name();
    my $dbh = $self->dbh();

    # Turn off autocommit so we can do all the deletes in one transaction.
    $dbh->{AutoCommit} = 0;

    foreach my $entry (@entries_to_delete) {
        my $qry = <<"EOQ";
DELETE FROM $table_name
WHERE id=$entry
EOQ
        $dbh->do($qry);
    }
    $dbh->commit();
    $dbh->{AutoCommit} = 1;

}

sub number_rows {
    my $self = shift();

    my $table_name = $self->table_name();
    my $qry = <<"EOQ";
SELECT COUNT(*) FROM $table_name
EOQ

    my $sth = $self->dbh()->prepare($qry);
    $sth->execute();

    my $number_of_rows;
    while (my @data = $sth->fetchrow_array()) {
        $number_of_rows = $data[0];
    }

    return $number_of_rows;
}

# Write an array of elements to the entriesDB.
sub write_array {
    my $self = shift();
    my ($elements_aref) = @_ ;

    my $table_name = $self->table_name();

    my @elements = @{ $elements_aref };

    my $dbh = $self->dbh();

    # Turn off autocommit so we can do all the inserts in one transaction.
    $dbh->{AutoCommit} = 0;

    my $sth = $self->dbh()->prepare("INSERT INTO $table_name(dn) VALUES (?)");
    foreach my $element (@elements) {
        $sth->bind_param(1, $element);
        $sth->execute();
    }
    $sth->finish();
    $dbh->commit();

    $dbh->{AutoCommit} = 1;

    return;
}


__PACKAGE__->meta->make_immutable;

1;
