package Stanford::LDAP::SyncScripts;

## no critic (Subroutines::RequireArgUnpacking) ;
## no critic (CodeLayout::ProhibitParensWithBuiltins) ;

use Net::LDAPapi ;
use YAML::Tiny ;
use Data::Dumper ;
use Net::LDAPapi ;
use Sys::Syslog ;
use Carp ;

use Stanford::LDAPserver;

# Make this a Moose class
use Moose ;
use namespace::autoclean ;

has 'debug' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has 'dryrun' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has 'start_time' => (
  is        => 'rw',
  isa       => 'Int',
  required  => 1,
  default   => time(),
) ;

has 'ldap' => (
  is        => 'rw',
  required  => 0,
  default   => undef,
) ;

has 'ldap_update' => (
  is        => 'rw',
  required  => 0,
  default   => undef,
) ;

has 'ldap_host' => (
  is        => 'rw',
  isa       => 'Maybe[Str]',
  required  => 1,
  default   => undef,
) ;

has 'principal_name' => (
  is        => 'rw',
  isa       => 'Maybe[Str]',
  required  => 0,
  default   => undef,
) ;

has 'keytab_file' => (
  is        => 'rw',
  isa       => 'Maybe[Str]',
  required  => 0,
  default   => undef,
) ;

has 'tgt_file' => (
  is        => 'rw',
  isa       => 'Maybe[Str]',
  required  => 0,
  default   => undef,
) ;

has 'log_base_dn' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default   => 'cn=accesslog',
);

has 'log_filter' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default => 'objectclass=*',
);

has 'log_cookie' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default => '/var/run/ldap-wg-maint.cookie',
);


### ### ### ### ### ### ### ### ### ### ### ### ### ###

sub dbg_message_format {
    my $self    = shift;
    my ($msg)   = @_;

    my $now     = time;
    my $elapsed = $now - $self->start_time();

    return "$now ($elapsed) $msg";
}

# Send debugging messages to STDOUT
sub dbg {
    my $self = shift;
    my ($msg) = @_;

    if ($self->debug()) {
        my $msg_formatted = $self->dbg_message_format($msg);
        print "$msg_formatted\n";
    }
    return;
}

# Generate informational and error messages.  Die if the severify is
# set to 'fatal'.
sub msg {
    my $self = shift;
    my ($severity, $msg) = @_;

    # syslog messages
    $severity = uc($severity);
    my $level;
    if ($severity eq 'FATAL') {
        $level = 'ERR';
    } else {
        $level = 'INFO';
    }
    my @lines = split(/\n/xsm, $msg);
    for my $l (@lines) { syslog($level, "$severity: $l"); }

    $self->dbg($severity . ': ' . $msg);

    # Stop processing if it is a bad error
    if ($severity eq 'FATAL') {
        croak($msg);
    } else {
        return;
    }
}

sub fatal {
    my $self = shift;
    my ($msg) = @_;

    return $self->msg('FATAL', $msg);
}

sub warning {
    my $self = shift;
    my ($msg) = @_;

    return $self->msg('WARNING', $msg);
}

sub format_ldap_error {
    my $self = shift;

    my ($ldap, $stat) = @_ ;

    my $m = q{};
    if ($ldap->errstring) {
        $m .= $ldap->errstring;
    }
    if ($ldap->extramsg) {
        $m .= ' - ' . $ldap->extramsg;
    }
    if (!$m) {
        $m = "Unknown LDAP error status = $stat";
    }

    return $m;
}

# We set up TWO LDAP handles: one mant for reading and one for for updating.
sub dir_connect {
    my $self = shift;

    $self->dbg('Binding to directory at ' . $self->ldap_host);

    my ($ldap_handle, $ldap_handle_update);

    $ldap_handle = ldap_connect(
        host      => $self->ldap_host,
        principal => $self->principal_name,
        keytab    => $self->keytab_file,
        tgt       => $self->tgt_file,
    );

    $ldap_handle_update = ldap_connect(
        host      => $self->ldap_host,
        principal => $self->principal_name,
        keytab    => $self->keytab_file,
        tgt       => $self->tgt_file,
    );

    $self->dbg('directory bind complete');

    $self->ldap($ldap_handle);
    $self->ldap_update($ldap_handle_update);

    return $ldap_handle;
}

# Return a hash result for an LDAP search.
sub search_ldap {
    my $self = shift;

    my ($ldap, $base, $scope, $filter, $attrs_aref, $context_msg) = @_;

    $self->dbg("ID Search: base:$base, filter:$filter, scope:$scope");

    my $stat = $ldap->search_s($base, $scope, $filter, $attrs_aref, 0);
    if ($stat == LDAP_NO_SUCH_OBJECT) {
        $self->warning("base:$base filter:$filter");
        $self->fatal("ERROR: $context_msg");
    } elsif ($stat != LDAP_SUCCESS) {
        $self->warning("LDAP search error base:$base, filter:$filter, scope:$scope");
        $self->fatal($self->format_ldap_error($ldap, $stat));
    }
    my $entry_ref = $ldap->get_all_entries;
    my %entry     = %$entry_ref;

    return %entry;
}

sub format_ldap_search {
    my $self = shift;
    my ($base, $scope, $filter, $attrs_aref, $context_msg) = @_;

    my $attributes_formatted = join("\n", @{ $attrs_aref }) ;

    my $result = <<"EOS";
base:      $base
scope:     $scope
filter:    $filter
attributes:
$attributes_formatted
EOS
    return $result;
}


sub populate_from_yaml_file {
    my $self = shift;
    my ($config_file) = @_;

    # Open the config
    my $yaml = YAML::Tiny->read($config_file);

    # Get a reference to the first document
    my $config = $yaml->[0];

    # Populate the common elements.
    my @attributes = $self->get_all_attribute_names();

    foreach my $attribute (@attributes) {
        if (exists($config->{$attribute})) {
            $self->$attribute($config->{$attribute});
        }
    }

    return ;
}

sub to_s {
    my $self = shift;

    my $spacer = '17';

    # Get all attributes.
    my @attributes = $self->get_all_attribute_names();

    my %attribute_to_constraint = $self->get_my_constraints();

    my $result = q{} ;
    foreach my $attribute (sort @attributes) {
        my $value = $self->$attribute ;
        if (!defined($value)) {
            $value = '<undef>';
        } elsif ($value eq q{}) {
            $value = '<empty string>';
        }

        my $constraint;
        if (exists $attribute_to_constraint{$attribute}) {
            $constraint = $attribute_to_constraint{$attribute};
        }

        if ($constraint && $constraint eq 'ArrayRef[Str]') {
            # Format an array
            $value = join("\n" . (q{ } x ($spacer + 2)), @{ $value });
        }

        if ($constraint && $constraint eq 'HashRef[Str]') {
            my @keys = sort keys %{ $value } ;
            my @formatted = ();
            foreach my $key (@keys) {
                my $v = $value->{$key};
                my $formatted = "$key => $v";
                push(@formatted, $formatted);

            }
            $value = join("\n" . (q{ } x ($spacer + 2)), @formatted);
        }

        if ($constraint && $constraint eq 'ArrayRef[HashRef[Str]]') {
            # Format an a hash of arrays
            my @values = @{ $value } ;
            my @f_results = ();
            foreach my $hash_ref (@{ $value }) {
                my @formatted = ();
                my %hash = %{ $hash_ref };
                foreach my $key (sort keys %hash) {
                    my $value1 = $hash{$key};
                    my $formatted = "$key => $value1";
                    push(@formatted, $formatted);
                }
                my $formatted_joined = join(', ', @formatted);
                push(@f_results, $formatted_joined );
            }
            $value = join("\n" . (q{ } x ($spacer + 2)), @f_results);
        }

        $result .= sprintf("%${spacer}s: %s", $attribute, $value) ;
        $result .= "\n";
    }
    return $result;
}

sub get_all_attribute_names {
    my $self = shift;

    return map { $_->name } $self->meta->get_all_attributes;
}

sub get_my_constraints {
    my $self = shift ;

    my %attribute_to_constraint = () ;

    my @attributes = $self->meta->get_all_attributes;
    foreach my $attribute (@attributes)  {
        if ($attribute->type_constraint) {
            $attribute_to_constraint{$attribute->name} = $attribute->type_constraint->name;
        }
    }

    return %attribute_to_constraint;
}
### ### ### ### ### ### ### ### ### ### ### ### ### ###


__PACKAGE__->meta->make_immutable;

1;
