package Stanford::LDAP::SyncScripts::WGMaint;

## no critic (Subroutines::RequireArgUnpacking) ;
## no critic (CodeLayout::ProhibitParensWithBuiltins) ;

use strict;
use warnings;
use autodie;

use Stanford::LDAPserver;
use Net::LDAPapi;
use Data::Dumper;

# Make this a Moose class extending Stanford::LDAP::SyncScripts
use Moose;

extends 'Stanford::LDAP::SyncScripts';

has 'force' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has 'failonwarn' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has 'select_attribute' => (
  is       => 'rw',
  isa      => 'Str',
  required => 1,
  default  => 'uid',
) ;

has 'map_attribute' => (
  is       => 'rw',
  isa      => 'Str',
  required => 1,
  default  => 'suSeasLocal',
) ;

has 'maps' => (
  is       => 'rw',
  isa      => 'ArrayRef[Str]',
  required => 1,
  default => sub { [] },
) ;

has 'ldap_base' => (
  is        => 'rw',
  isa       => 'Maybe[Str]',
  required  => 0,
  default   => 'cn=accounts,dc=stanford,dc=edu',
) ;

has 'ldap_filter' => (
  is       => 'rw',
  isa      => 'Str',
  required => 1,
  default  => '(suSeasLocal=*)',
) ;

has 'ldap_remove_filter' => (
  is       => 'rw',
  isa      => 'Str',
  required => 1,
  default  => '(&(!(suSeasLocal=*))(suPrivilegeGroup=itservices:mail-*))',
) ;

has 'webservice_host' => (
  is       => 'rw',
  isa      => 'Str',
  required => 1,
  default  => 'workgroupsvc-dev.stanford.edu',
) ;

has 'webservice_version' => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => 'v1',
) ;

has 'webservice_cert' => (
  is       => 'rw',
  isa      => 'Str',
  required => 1,
  default  => '/etc/ssl/certs/wg-api.ldap-dev.stanford.edu.pem',
) ;

has 'webservice_key' => (
  is       => 'rw',
  isa      => 'Str',
  required => 1,
  default  => '/etc/ssl/private/wg-api.ldap-dev.stanford.edu.key',
) ;

has 'wg_regex' => (
  is       => 'rw',
  isa      => 'ArrayRef[HashRef[Str]]',
  required => 1,
  default => sub { [] },
) ;

has 'wg_valid' => (
  is       => 'rw',
  isa      => 'HashRef[Str]',
  required => 1,
  default => sub { {} },
) ;

sub populate_maps {
    my $self = shift ;

    # Put the list of valid workgroups into a hash for later use.
    my $map_aref = $self->maps();
    if (scalar(@{$map_aref}) == 0) {
        $self->fatal('No map entries specified in the configuration file');
    }

    my %wg_valid = ();
    my @wg_regex = ();
    for my $map_entry (@{$map_aref}) {
        my ($regex, $wg) = split(m{=}xsm, $map_entry);
        $wg    =~ s{\s+}{}xmsg;
        $regex =~ s{^\s+}{}xmsg;
        $regex =~ s{\s+$}{}xmsg;
        $wg_valid{$wg}++;
        push(@wg_regex, { regex => $regex, wg => $wg });
    }

    $self->wg_regex(\@wg_regex);
    $self->wg_valid(\%wg_valid);

    return;
}

# ------------------------------------------------------------------------
# Take an attribute value and map it to a workgroup.
#
# For example, if the settings in the configuration file are:
#
#   map = .*?@zm[012]\d.stanford.edu = itservices:mail-zimbra
#   map = .*?@zm88.stanford.edu = itservices:mail-gmail
#   map = .*?@windows.stanford.edu = itservices:mail-onprem-exchange
#   map = .*?@office365stanford.mail.onmicrosoft.com = itservices:mail-office365
#   map = .*?@o365.stanford.edu = itservices:mail-office365
#
# then $val will map to one of the itservices:* values.
sub map_to_workgroup {
    my $self = shift ;
    my ($val) = @_;

    my $return_wg;
    for my $this_map_ref (@{ $self->wg_regex }) {
        my %this_map = %{$this_map_ref};
        if ($val =~ /$this_map{'regex'}/xms) {
            $return_wg = $this_map{'wg'};
            last;
        }
    }
    return $return_wg;
}

# Given a base dn return the select_attribute's value (probably uid).
sub get_id {
    my $self = shift ;
    my ($this_dn) = @_;

    my $sel_attr = $self->select_attribute;
    my $scope    = LDAP_SCOPE_BASE;
    my $base     = $this_dn;
    my @attrs    = ($sel_attr);
    my $filter   = 'objectclass=*';

    my $context_msg = "Problem searching for $this_dn";
    my %entry = $self->search_ldap($self->ldap_update,
                                   $base,
                                   $scope,
                                   $filter,
                                   \@attrs,
                                   $context_msg);
    my $entry_cnt = keys %entry;
    if (!$entry_cnt) {
        return;
    }

    my $val;
    for my $dn (keys %entry) {
        my $cnt = 0;
        for my $a (keys %{ $entry{$dn} }) {
            if (lc($a) eq lc($self->select_attribute)) {
                for my $v (@{ $entry{$dn}{$a} }) {
                    $val = $v;
                }
                last;
            }
        }
    }
    return $val;
}

# ------------------------------------------------------------------------
# Decode webservice XML error return
#
# The web service returns errors as XML documents of the form:
#
#  <?xml version="1.0" encoding="UTF-8" ?>
#  <error>
#  <code>409</code>
#  <message>person/whm  already exist in itservices:mail-zimbra</message>
#  </error>

sub ws_decode_error {
    my $self = shift ;
    my ($s) = @_;

    my @fld_list = qw(code message);
    my %val = ();
    for my $fld (@fld_list) {
        if ($s =~ /\<$fld\> (.+?) \<\/$fld\>/xms) {
            $val{$fld} = $1;
        }
    }

    return %val;
}

# ------------------------------------------------------------------------
# Check the member list for all of the Workgroups.  Add and delete as
# required.
#
# If a user's sunetid is passed sync only that user; otherwise, sync
# EVERYONE who has an suSeasLocal attribute .
#
# Unless the defaults have changed this means given a user's sunetid
# find

# 1. Given a user's sunetid get that user's suSeasLocal attribute
#    (map_attribute) and all of the user's suPrivilegeGroup attributes.

# 2. If the user does not have an suSeasLocal attribute we stop.

# 3. If suSeasLocal does not map to one of the itservices:mail-* groups
#    give up.

# 4. Otherwise, we know that suSeasLocal maps to one of the
#    itservices:mail-* groups.

# 5A. The user does not belong to the corresponding itservices:mail-* group, so ADD
#     that user to the group.

# 5B. The user belongs to an itservices:mail-* group that is not the correct one, so
#     REMOVE that user from that group.

sub sync_workgroups {
    my $self = shift ;
    my ($this_id, $opt_force) = @_;

    if ($this_id) {
        $self->dbg("starting sync_workgroups for '$this_id'");
    } else {
        $self->dbg('starting sync_workgroups for all ids');
    }

    my $map_attr = $self->map_attribute;
    my $sel_attr = $self->select_attribute;
    my $scope    = LDAP_SCOPE_SUBTREE;
    my $base     = $self->ldap_base;
    my @attrs    = ($map_attr, $sel_attr, 'suPrivilegeGroup');

    # Create the filter
    my $filter = q{};
    if ($this_id) {
        # CASE 1. Sync only the user matching $this_id.
        # A user's id was passed so we restrict to just this user.
        $filter .= "(&(${sel_attr}=${this_id})";
        if ($self->ldap_filter =~ /^[(]/xms) {
            $filter .= $self->ldap_filter;
        } else {
            $filter .= '(' . $self->ldap_filter . ')';
        }
        $filter .= ')';
    } else {
        # CASE 2. Sync everyone.
        $filter = $self->ldap_filter;
    }

    $self->dbg("using filter: $filter");

    my $context_msg = 'Problem searching group branch';
    my %entry = $self->search_ldap($self->ldap_update,
                                   $base,
                                   $scope,
                                   $filter,
                                   \@attrs,
                                   $context_msg);

    my $entry_cnt = keys %entry;
    if ($self->debug) {
        $self->dbg("entry_cnt:$entry_cnt");
    }

    if (!$entry_cnt) {
        return;
    }

    for my $dn (keys %entry) {
        my $cnt = 0;

        # Pulling the attributes using case insensitive tests
        my $attr_value;
        my $id;
        for my $a (%{ $entry{$dn} }) {
            if (lc($a) eq lc($self->map_attribute)) {
                for my $v (@{ $entry{$dn}{$a} }) {
                    $attr_value = $v;
                }
            } elsif (lc($a) eq lc($self->select_attribute)) {
                for my $v (@{ $entry{$dn}{$a} }) {
                    $id = $v;
                }
            }
        }
        if (!$id || !$attr_value) {
            $self->dbg("attr_value:$attr_value id:$id");
            $self->dbg("Skipping $dn due to missing value or id");
            next;
        }

        # Check to make sure the user is a member of the correct work
        # group.
        my $this_wg = $self->map_to_workgroup($attr_value);
        if (!$this_wg) {
            $self->dbg('No mapping for '
                              . $self->map_attribute . ': '
                              . $attr_value);
            next;
        }
        my @work_add = ();
        my @work_del = ();
        my $found    = 0;
        $self->dbg("$attr_value -> $this_wg");
        for my $val (@{ $entry{$dn}{'suPrivilegeGroup'} }) {
            my $test_val = lc($val);
            if ($test_val eq $this_wg) {
                # It appears that this user is already a member
                # of the correct itservices:mail-* group.
                $found++;
            } elsif ($self->wg_valid->{$test_val}) {
                # $test_val is one of the itservices:mail-* groups
                # but does not agree with $this_wg, so the user must
                # be deleted from the $test_val group.
                push(@work_del, $val);
            }
        }
        if (!$found || $opt_force) {
            push(@work_add, $this_wg);
        }
        if (scalar(@work_add)) {
            $self->add_members($id, \@work_add, $self->dryrun);
        }
        if (scalar(@work_del)) {
            $self->del_members($id, \@work_del);
        }
    }

    $self->ldap_update->msgfree;

    return;
}

# ------------------------------------------------------------------------
# Check workgroups for members that need to be removed

sub check_workgroups {
    my $self = shift ;

    my $map_attr = $self->map_attribute;
    my $sel_attr = $self->select_attribute;
    my $scope    = LDAP_SCOPE_SUBTREE;
    my $base     = $self->ldap_base;
    my @attrs    = ($map_attr, $sel_attr, 'suPrivilegeGroup');
    my $filter   = $self->ldap_remove_filter;

    my $ldap_update = $self->ldap_update;

    $self->dbg("Check Search. base:$base filter:$filter scope:$scope");
    my $stat = $ldap_update->search_s($base, $scope, $filter, \@attrs, 0);
    if ($stat == LDAP_NO_SUCH_OBJECT) {
        $self->msg('warning', "base:$base filter:$filter");
        $self->msg('fatal',   'ERROR: Problem searching group branch');
    } elsif ($stat != LDAP_SUCCESS) {
        $self->msg('warning',
            "LDAP search error Base:$base Filter:$filter Scope:$scope");
        my $m;
        if ($ldap_update->errstring) {
            $m .= $ldap_update->errstring;
        }
        if ($ldap_update->extramsg) {
            $m .= ' - ' . $ldap_update->extramsg;
        }
        if (!$m) {
            $m = "Unknown LDAP error status = $stat"
        }
        $self->msg('fatal', $m);
    }
    my $entry_ref = $ldap_update->get_all_entries;
    my %entry     = %$entry_ref;
    my $entry_cnt = keys %entry;
    $self->dbg("entry_cnt:$entry_cnt");
    if (!$entry_cnt) {
        return;
    }

    for my $dn (keys %entry) {
        my $cnt = 0;

        # Use a hash for the workgroups to remove list so that
        # duplicates are handled automatically.
        my %del_hash = ();
        my $id;
        for my $a (%{ $entry{$dn} }) {
            if (lc($a) eq 'suprivilegegroup') {
                for my $v (@{ $entry{$dn}{$a} }) {
                    if ($self->wg_valid->{$v}) {
                        $del_hash{$v}++;
                    }
                }
            } elsif (lc($a) eq lc($self->select_attribute)) {
                for my $v (@{ $entry{$dn}{$a} }) {
                    $id = $v;
                }
            }
        }
        if (!$id || keys(%del_hash) == 0) {
            my @key_list = keys %del_hash;
            $self->dbg('suPrivilegeGroups:' . join(q{ }, @key_list) . " id:$id");
            $self->dbg("Skipping $dn due to missing workgroups or id");
            next;
        }
        # Create the array to pass to the delete routine
        my @del_list = ();
        for my $wg (keys %del_hash) {
            push @del_list, $wg;
        }
        del_members($id, \@del_list);
    }

    $ldap_update->msgfree;
    return;
}

# ------------------------------------------------------------------------
# Add a member to a list of workgroups
#
#  url:    https://workgroupsvc.stanford.edu/v1/workgroups/{name}/members
#  method: put
#  string: user=https://workgroupsvc.stanford.edu/v1/users/{sunetid}

sub add_members {
    my $self = shift;
    my ($id, $work_ref) = @_;

    my $ua = LWP::UserAgent->new();
    $ua->timeout(30);

    for my $wg (@{$work_ref}) {
        $self->dbg("Adding id:$id to wg:$wg");
        my $url = $self->make_user_url($wg, $id);
        if ($self->dryrun) {
            $self->dbg("Proposing add: $url");
            next;
        }
        $self->dbg("url: $url");
        my $response = $ua->put($url);
        if ($response->is_success) {
            $self->msg('INFO', "$id added to Workgroup $wg");
        } else {
            my %err = $self->ws_decode_error($response->decoded_content);
            $self->msg('WARN', $err{'code'} . ' - ' . $err{'message'});
            $self->msg('WARN', $response->status_line);
            if ($self->failonwarn) {
                $self->msg('FATAL', 'stopping processing at add_members');
            }
        }
    }
    return;
}

# ------------------------------------------------------------------------
# Make a sure to hand to ADD and DELETE Workgroup web services

sub make_user_url {
    my $self = shift;
    my ($wg, $id) = @_;

    my $url = 'https://' . $self->webservice_host;
    $url .= q{/} . $self->webservice_version;
    $url .= '/workgroups/' . $wg . '/members';

    my $qstr = 'user=https://' . $self->webservice_host;
    $qstr .= q{/} . $self->webservice_version . "/users/$id";

    return "${url}?${qstr}";
}

# ------------------------------------------------------------------------
# Delete a member from a list of workgroups
#
#  url:    https://workgroupsvc.stanford.edu/v1/workgroups/{name}/members
#  method: delete
#  string: user=https://workgroupsvc.stanford.edu/v1/users/{sunetid}

sub del_members {
    my $self = shift;
    my ($id, $work_ref) = @_;

    my $ua = LWP::UserAgent->new();
    $ua->timeout(30);

    for my $wg (@{$work_ref}) {
        $self->dbg("Removing id:$id from wg:$wg");
        my $url = $self->make_user_url($wg, $id);
        if ($self->dryrun) {
            $self->dbg("Proposing delete: $url");
            next;
        }
        $self->dbg("url: $url");
        my $response = $ua->delete($url);
        if ($response->is_success) {
            $self->msg('INFO', "$id removed from Workgroup $wg");
        } else {
            my %err = $self->ws_decode_error($response->decoded_content);
            $self->msg('WARN', $err{'code'} . ' - ' . $err{'message'});
            $self->msg('WARN', $response->status_line);
            if ($self->failonwarn) {
                $self->msg('FATAL', 'stopping processing at del_members');
            }
        }
    }
    return;
}

# ------------------------------------------------------------------------
# listen for any changes published in the accesslog

sub listen_for_changes {
    my $self = shift;

    my $this_base   = $self->log_base_dn;
    my $this_filter = $self->log_filter();
    my $this_cookie = $self->log_cookie();

    # Make sure there is a cookie file
    if (!-e $this_cookie) {
        my $f;
        open($f, '>', $this_cookie);
        close $f;
    }

    my $ldap = $self->ldap;

    my $msgid = $ldap->listen_for_changes(
        -basedn => $this_base,
        -scope  => LDAP_SCOPE_SUBTREE,
        -filter => $this_filter,
        -cookie => $this_cookie
    );

    my $base_dn  = squeeze($self->ldap_base);
    my $map_attr = $self->map_attribute;

    # Loop forever, listening for changes.
    while (1) {
        $self->dbg('listen loop');
        while (my @entries = $ldap->next_changed_entries($msgid, 0, -1)) {
            foreach my $entry (@entries) {

                my $sync_entry = 0;

                # Find the dn of whatever changed
                my @reqdn = ();
                if (!(@reqdn = $ldap->get_values('reqDN'))) {
                    next;
                }
                my $dn = squeeze($reqdn[0]);
                $self->dbg("change dn:$dn");
                if ($dn !~ /$base_dn$/xmsi) {
                    next;
                }

                # Look for map_attribute changes
                my @mods = ();
                if (!(@mods = $ldap->get_values('reqMod'))) {
                    next;
                }
                foreach my $val (@mods) {
                    if ($val =~ /$map_attr: \s* (.*)/xms) {
                        $self->dbg("mapping $map_attr of $dn");
                        my $id = $self->get_id($dn);
                        if ($id) {
                            $self->sync_workgroups($id);
                        } else {
                            $self->msg('warn',
                                $self->select_attribute
                                  . " not found for $dn");
                        }
                    }
                }
            }

            # Free the current LDAP result. This should help with any
            # memory leaks.
            $ldap->msgfree;
        }
    }
    return;
}



1;
__END__
