package Stanford::LDAP::SyncSuGAL;

## no critic (RequireArgUnpacking);

use Carp;
use Data::Dumper;
use AppConfig qw(:argcount :expand);
use Log::Log4perl;
use YAML::Tiny;
use Authen::Krb5;
use Net::LDAPapi;
use File::Basename;
use Time::HiRes qw/usleep/;

use Stanford::LDAPserver;
use Stanford::LDAP::EntriesDB;

use strict;
use warnings;
use autodie;

# Make this a Moose class
use Moose ;
use namespace::autoclean;

has 'update' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

# The amout of time, in milliseconds, to pause after each synchronization
# attempt. The default is zero.
has 'pause_milliseconds' => (
  is        => 'rw',
  isa       => 'Int',
  required  => 1,
  default   => 0,
) ;

has 'is_listener' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has 'debug' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has 'log_spacer' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default   => q{},
) ;

has 'log_indent' => (
  is        => 'rw',
  isa       => 'Int',
  required  => 1,
  default   => 0,
) ;

has 'retry_count' => (
  is        => 'rw',
  isa       => 'Int',
  required  => 1,
  default   => 5,
) ;

# The master config file:
my $CONFIG_FILE_DEFAULT = '/etc/ldapadmin/ldap-sync-sugal-attributes.conf';
has 'config_file' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default   => $CONFIG_FILE_DEFAULT,
) ;

has 'visibility_attributes_file' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default   => '/etc/ldapadmin/sugal-visibility.yaml',
) ;

# The logger object.
has 'logger'  => (
  is          => 'rw',
  isa         => 'Log::Log4perl::Logger',
  required    => 0,
) ;

has 'log4perl_config'  => (
  is          => 'rw',
  isa         => 'Str',
  required    => 1,
  default     => q{},
) ;

# If this is set to 1, then log messages that might contain private
# directory information. The default is to set this to 0.
has 'log_private' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

# The logger object.
has 'CONF'  => (
  is        => 'rw',
  isa       => 'AppConfig',
  required    => 0,
) ;

# Master hash mapping visibility attributes to the attributes they
# protect.
has 'VISIB_ATTRIBUTES' => (
  is        => 'rw',
  isa       => 'HashRef',
  required  => 1,
  default   => sub { {} },
) ;

# A hash mapping a person attribute to the visibility attribute that
# controls it.
has 'ATTR_TO_VISIB' => (
  is        => 'rw',
  isa       => 'HashRef',
  required  => 1,
  default   => sub { {} },
) ;

# A hash mapping the person attribute to its corresponding suGAL
# attribute.
has 'SRCATTR_TO_TARGATTR' => (
  is        => 'rw',
  isa       => 'HashRef',
  required  => 1,
  default   => sub { {} },
) ;

# The DN of the base of the people tree.
has 'people_dn' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 1,
  default   => 'cn=people,dc=stanford,dc=edu',
) ;

# Should be one of 'accounts' or 'peeople'
has 'base_tree_filter' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;

has 'ldap_host' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;

has 'principal_name' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;

has 'keytab_file' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;

has 'tgt_file' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;

# Ignore keytabs and tgt files and just depend on the KRB5CCNAME
# environment variable for Kerberos credentials.
has 'use_krb5ccname' => (
  is        => 'rw',
  isa       => 'Bool',
  required  => 1,
  default   => 0,
) ;

has 'LDAP' => (
  is        => 'rw',
  isa       => 'Any',
  required  => 1,
  default  => undef,
) ;

has 'LDAP_UPDATE' => (
  is        => 'rw',
  isa       => 'Any',
  required  => 1,
  default  => undef,
) ;

has 'in_filter' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;

has 'log_base_dn' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;


has 'log_cookie_file' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;


has 'log_filter' => (
  is        => 'rw',
  isa       => 'Str',
  required  => 0,
) ;

# Limit the number of entries UPDATED when using the sync command to this
# number. Has no effect when running in listen mode. The default is 0
# meaning no limit.
has 'limit' => (
  is        => 'rw',
  isa       => 'Int',
  required  => 0,
  default   => 0,

) ;

# When this object goes away, delete the tgt file.
sub DEMOLISH {
    my $self = shift;
    if ($self->tgt_file() && (-f $self->tgt_file())) {
        unlink $self->tgt_file();
    }
    return;
}

our $VERSION = '0.01' ;


## Logging-related code

## WARNING! WARNING! WARNING! WARNING!
#
# Be careful that no actual attribute values get logged as this might
# expose directory data marked private into a log server that a wider
# audience has access to.
#
## WARNING! WARNING! WARNING! WARNING!

sub logger_init {
    my $self = shift;

    my $log4perl_config = $self->log4perl_config() ;

    # If no Log4Perl configuration file passed, log INFO and above to
    # syslog.
    if (!$log4perl_config) {
        my %key_value_pairs = (
            'log4perl.rootLogger'                => 'INFO, SYSLOG',
            'log4perl.appender.SYSLOG'           => 'Log::Dispatch::Syslog',
            'log4perl.appender.SYSLOG.layout'    => 'SimpleLayout',
        );

        Log::Log4perl->init( \%key_value_pairs );
    } elsif (!(-e $log4perl_config)) {
        croak "could not load Log4Perl configuration file '$log4perl_config'";
    } else {
        Log::Log4perl->init($log4perl_config);
    }

    my $logger = Log::Log4perl->get_logger('Stanford::LDAP::SyncSuGAL');
    $self->logger($logger);

    return $self->logger();
}

sub log_prep_msg {
    my $self = shift;
    my ($msg) = @_;

    my $spacer = $self->log_spacer();
    my $log_indent = $self->log_indent();

    my $listener_suffix;
    if ($self->is_listener()) {
        $listener_suffix = ' (listener)';
    } else {
        $listener_suffix = q{};
    }
    return ($spacer x $log_indent) . $msg . $listener_suffix ;
}

sub log_fatal {
    my $self = shift;
    my ($msg) = @_ ;

    $msg = $self->log_prep_msg($msg);
    $self->logger()->fatal($msg);
    croak "FATAL error: $msg\nExiting";
}

sub log_error {
    my $self = shift;
    my ($msg) = @_ ;

    $msg = $self->log_prep_msg($msg);
    $self->logger()->error($msg);
    return;
}

sub log_warn {
    my $self = shift;
    my ($msg) = @_ ;

    $msg = $self->log_prep_msg($msg);
    $self->logger()->warn($msg);
    return;
}

sub log_info {
    my $self = shift;
    my ($msg) = @_ ;

    $msg = $self->log_prep_msg($msg);
    $self->logger()->info($msg);
    return;
}

sub log_debug {
    my $self = shift;
    my ($msg) = @_ ;
    if (!defined($msg)) {
        $msg = '<undef>';
    }
    if ($self->debug()) {
        print "$msg\n";
    }

    $msg = $self->log_prep_msg($msg);
    $self->logger()->debug($msg);
    return;
}

sub log_trace {
    my $self = shift;
    my ($msg) = @_ ;

    $msg = $self->log_prep_msg($msg);
    $self->logger()->trace($msg);
    return;
}

sub log_indent_inc {
    my $self = shift;

    my $log_indent = $self->log_indent();

    $self->log_indent($log_indent + 1);
    return;
}

sub log_indent_dec {
    my $self = shift;

    my $log_indent = $self->log_indent();

    if ($log_indent) {
        $self->log_indent($log_indent - 1);
    }
    return;
}

#sub update_pause {
#    my $self = shift;
#
#    if ($self->update_pause_milliseconds() > 0) {
#        # Sleep this many microseconds.
#        usleep(1000 * $self->update_pause_milliseconds());
#    }
#    return;
#}

sub pause {
    my $self = shift;

    if ($self->pause_milliseconds() > 0) {
        # Sleep this many microseconds.
        usleep(1000 * $self->pause_milliseconds());
    }
    return;
}

# returns an AppConfig object.
sub read_config_file_into_appconfig {
    my $self = shift;

    my ($config_file) = @_ ;

    $self->config_file($config_file);

    my($filename, $dirs, $suffix) = fileparse($0);

    my $APPCONF = AppConfig->new({});

    $APPCONF->define('ldap_host',
                  {ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('visibility_attributes_file',
                  {ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('people_dn',
                  {DEFAULT  => 'cn=people,dc=stanford,dc=edu',
                   ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('base_tree_filter',
                  {DEFAULT  => 'accounts',
                   ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('in_filter',
                  {DEFAULT  => 'objectClass=*',
                   ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('log_base_dn',
                  {DEFAULT  => 'cn=accesslog',
                   ARGCOUNT => ARGCOUNT_ONE});
    $APPCONF->define('log_cookie_file',
                  {DEFAULT  => '/var/run/ldap-sync-attribute.cookie',
                   ARGCOUNT => ARGCOUNT_ONE});
    $APPCONF->define('log_filter',
                  {DEFAULT  => 'objectclass=*',
                   ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('retry_count',
                  {DEFAULT  => 5,
                   ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('keytab_file',
                  {ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('principal_name',
                  {ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('tgt_file',
                  {ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('remctl_subcommand',
                  {DEFAULT  => 'privsync',
                   ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('debug',
                  {DEFAULT  => 0,
                   ARGCOUNT => ARGCOUNT_NONE});


    $APPCONF->define('log4perl_config',
                  {DEFAULT  => q{},
                   ARGCOUNT => ARGCOUNT_ONE});

    $APPCONF->define('update',
                  {DEFAULT  => 0,
                   ARGCOUNT => ARGCOUNT_NONE});

    $APPCONF->define('use_krb5ccname',
                  {DEFAULT  => 0,
                   ARGCOUNT => ARGCOUNT_NONE});

    $APPCONF->file($config_file);

    # Replicate to $self.
    $self->ldap_host($APPCONF->ldap_host);
    $self->base_tree_filter($APPCONF->base_tree_filter());
    $self->people_dn($APPCONF->people_dn());

    # Don't try to set these attributes if the $APPCONF value is undefined
    # or empty.
    if ($APPCONF->keytab_file) {
        $self->keytab_file($APPCONF->keytab_file);
    }

    if ($APPCONF->principal_name) {
        $self->principal_name($APPCONF->principal_name);
    }

    # The TGT file.
    #
    # Sometimes we run with an existing KRB5CCNAME environment variable
    # set, so we don't always need a tgt_file.

    if ($APPCONF->tgt_file) {
        $self->tgt_file($APPCONF->tgt_file);
    }

    # If the keytab_file is defined but tgt_file is NOT defined,
    # set a default tgt_file.
    if ($self->keytab_file() && !$self->tgt_file()) {
        my $default_tgt = '/tmp/krb5_'. $filename .'_'.$$.'tgt';
        $self->tgt_file($default_tgt);
    }

    $self->in_filter($APPCONF->in_filter());
    $self->retry_count($APPCONF->retry_count());

    $self->log_base_dn($APPCONF->log_base_dn());
    $self->log_cookie_file($APPCONF->log_cookie_file());
    $self->log_filter($APPCONF->log_filter());

    $self->log4perl_config($APPCONF->log4perl_config());

    $self->visibility_attributes_file($APPCONF->visibility_attributes_file());

    $self->use_krb5ccname($APPCONF->use_krb5ccname());



    return $APPCONF;
}

# Read the configuration file and populate some global data structures:
#   $CONF
#   %VISIB_ATTRIBUTES;
#   %ATTR_TO_VISIB;
#   %SRCATTR_TO_TARGATTR;
#
# arguments: pass a hash reference
#
#   {
#      'conf_file' => full path to configuration file
#      'overrides' => a hash reference of configuration file overrides
#      'skip_checks' => a boolean that when true indicates that most
#                       checks shoudl be skipped (for package building).
#   }
sub read_config {
    my $self = shift;

    my ($arg_href) = @_;

    # Figure out what the configuration file is.
    my $config_file =
        $arg_href->{'config_file'} ? $arg_href->{'config_file'} : $CONFIG_FILE_DEFAULT;
    my $overrides_href =
        $arg_href->{'overrides'}   ? $arg_href->{'overrides'}   : {};
    my $skip_checks =
        $arg_href->{'skip_checks'} ? $arg_href->{'skip_checks'} : 0;

    if (!$config_file) {
        croak 'no configuration file found';
    } elsif (!(-e $config_file)) {
        croak "cannot find configuration file '$config_file'";
    }

    # Sets many $self attributes.
    my $APPCONF = $self->read_config_file_into_appconfig($config_file);

    # Initialize the logger
    $self->logger_init($self->log4perl_config());

    # Make sure the ldap_host is valid.
    if (!$skip_checks && !$self->valid_host($APPCONF->ldap_host())) {
        $self->log_fatal(q{Invalid host name '} . $APPCONF->ldap_host() . q{'});
    }

    # Process overrides
    if (exists($overrides_href->{'debug'})) {
        $self->debug($overrides_href->{'debug'});
    }

    if (exists($overrides_href->{'keytab_file'})) {
        $self->keytab_file($overrides_href->{'keytab_file'});
    }

    if (exists($overrides_href->{'principal_name'})) {
        $self->principal_name($overrides_href->{'principal_name'});
    }

    if (exists($overrides_href->{'in_filter'})) {
        $self->in_filter($overrides_href->{'in_filter'});
    }

    if (exists($overrides_href->{'update'})) {
        $self->update($overrides_href->{'update'});
    }

    if (exists($overrides_href->{'limit'})) {
        $self->limit($overrides_href->{'limit'});
    }

    if (exists($overrides_href->{'pause_milliseconds'})) {
        $self->pause_milliseconds($overrides_href->{'pause_milliseconds'});
    }


    # More error checking

    # Make sure the keytab file exists.
    if (!$skip_checks && $APPCONF->keytab_file && (!(-e $APPCONF->keytab_file))) {
        $self->log_fatal('Keytab file not found (' . $APPCONF->keytab_file . q{)})
    }

    # If a keytab file is specified then require a principal name
    if (!$skip_checks && $APPCONF->keytab_file && (!$APPCONF->principal_name)) {
        $self->log_fatal('ERROR: principal name required if keytab specified');
    }

    # Read the visibility YAML file and set the related hash references.
    $self->read_and_set_visibility_attributes();

    $self->log_trace("finished configuration file '$config_file' parse");

    $self->CONF($APPCONF);

    return;
}

sub config_string {
    my $self = shift;

    my @settings = qw/
      config_file
      log4perl_config
      ldap_host
      base_tree_filter
      in_filter
      update
      log_base_dn
      log_cookie_file
      log_filter
      use_krb5ccname
      principal_name
      keytab_file
      retry_count
      visibility_attributes_file
    /;

    foreach my $setting (@settings) {
        my $value = $self->{$setting};
        if (!defined($value)) {
            $value = '<UNDEF>';
        }
        print sprintf("%-17s %s\n", $setting . q{:}, $value);
    }
    return;
}

sub valid_host {
    my $self = shift;

    my ($hn) = @_;
    my @ip = split(/\./xsm, $hn);

    # check for a ip literal
    if (scalar(@ip) == 4) {
        my $okay = 1;
        for my $i (0..3) {
            if ($ip[$i] !~ m{^\d+$}xsm) {
                $okay = 0;
            }
            if ($ip[$i] < 0) {
                $okay = 0;
            }
            if ($ip[$i] > 255) {
                $okay = 0;
            }

            if (!$okay) {
                last;
            }
        }
        return 1 if $okay;
    }

    # If we don't get an octets it is because it is a bad hostname
    my $raw_addr = (gethostbyname($hn))[4];
    my @octets = unpack('C4', $raw_addr);
    if (!scalar(@octets)) {
        return 0;
    }

    # Don't use the hostname, but might need it sometime
    my $ip = join(q{.}, @octets);
    return 1;
}

# Return the number of entries given a search base and filter

sub get_entry_cnt {
    my $self = shift;

    my $LDAP = $self->LDAP();
    my ($base, $filter, $scope) = @_;

    my @attrs = ('objectclass');
    my $stat = $LDAP->search_s($base,$scope,$filter,\@attrs,0);
    return 0 if $stat == LDAP_NO_SUCH_OBJECT;
    if ($stat != LDAP_SUCCESS) {
        msg ('warning',
             "LDAP search error Base:$base Filter:$filter Scope:$scope");
        my $m;
        if ($LDAP->errstring) {
            $m .= $LDAP->errstring;
        }
        if ($LDAP->extramsg) {
            $m .= ' - '.$LDAP->extramsg;
        }
        if (!$m) {
            $m = "Unknown LDAP error status = $stat";
        }
        msg ('fatal', $m);
    }
    my $entry_ref = $LDAP->get_all_entries;
    my %entry = %{ $entry_ref };
    my $entry_cnt = keys %entry;
    return $entry_cnt;
}

# Read the YAML file that contains the visibility attributes with the set
# of attributes they protect.
#
# Sets these properties:
#   $self->VISIB_ATTRIBUTES
#   $self->ATTR_TO_VISIB
#   $self->SRCATTR_TO_TARGATTR
#
#
# {
#    'suVisibIdentity' ->
#      {
#          'cn'          -> 'suGALcn',
#          'displayName' -> 'suGALdisplayName',
#          ...
#      },
#    'suVisibEmailPager' ->
#      {
#          'emailPager' -> 'suGALEmailPager',
#      },
#    ...
# }
#
# Sets these properties:
#   $self->VISIB_ATTRIBUTES
#   $self->ATTR_TO_VISIB
#   $self->SRCATTR_TO_TARGATTR

sub read_and_set_visibility_attributes {
    my $self = shift;
    my $visib_file = $self->visibility_attributes_file();

    if (!(-e $visib_file)) {
        $self->log_fatal("cannot find visibility mapping file '$visib_file'");
    }
    # Read this YAML file.

    # Open the config
    my $yaml = YAML::Tiny->read($visib_file);
    $self->log_debug("read the YAML file '$visib_file'.");

    # Get a reference to the first document
    my %VISIB_ATTRIBUTES = %{ $yaml->[0] };
    $self->log_trace('extracted the config from the YAML object:');

    $self->VISIB_ATTRIBUTES(\%VISIB_ATTRIBUTES);

    # Populate %ATTR_TO_VISIB and %SRCATTR_TO_TARGATTR
    my %ATTR_TO_VISIB       = ();
    my %SRCATTR_TO_TARGATTR = ();
    $self->log_trace('about to populate \%ATTR_TO_VISIB and \%SRCATTR_TO_TARGATTR');
    foreach my $visib_attr (keys %VISIB_ATTRIBUTES) {
        my $attribute2gal_href = $VISIB_ATTRIBUTES{$visib_attr};
        my @reg_attributes = keys %{$attribute2gal_href};
        foreach my $reg_attribute (@reg_attributes) {
            if (exists($ATTR_TO_VISIB{$reg_attribute})) {
                $self->log_fatal("attribute $reg_attribute is controlled by two"
                                 . "visibility attributes: $visib_attr and "
                                 . $ATTR_TO_VISIB{$reg_attribute});
            }
            $ATTR_TO_VISIB{$reg_attribute} = $visib_attr;
        }
        %SRCATTR_TO_TARGATTR = (%SRCATTR_TO_TARGATTR, %{ $attribute2gal_href });
    }

    $self->ATTR_TO_VISIB(\%ATTR_TO_VISIB);
    $self->SRCATTR_TO_TARGATTR(\%SRCATTR_TO_TARGATTR);

    return;
}

sub bind_ldap_reader {
    my $self = shift;

    $self->log_debug($self->ldap_host()) ;
    $self->log_debug($self->principal_name()) ;
    $self->log_debug($self->keytab_file()) ;
    $self->log_debug($self->tgt_file()) ;

    my $LDAP;
    if ($self->use_krb5ccname()) {
        $self->log_debug('(bind_ldap_reader) using KRB5CCNAME instead of keytab/tgt') ;
        $LDAP = ldap_connect(host => $self->ldap_host(),);
    } else {
        $LDAP = ldap_connect(host => $self->ldap_host(),
                             principal => $self->principal_name(),
                             keytab    => $self->keytab_file(),
                             tgt       => $self->tgt_file(),
            );
    }

    $self->log_debug('bind complete');
    $self->log_debug('bound to host: ' . $self->ldap_host()) ;

    $self->LDAP($LDAP);
    return $self->LDAP();
}

sub unbind_ldap_reader {
    my $self = shift;
    return ldap_disconnect($self->LDAP());
}

sub bind_ldap_writer {
    my $self = shift;

    my $LDAP_UPDATE;
    if ($self->use_krb5ccname()) {
        $self->log_debug('(bind_ldap_writer) using KRB5CCNAME instead of keytab/tgt') ;
        $LDAP_UPDATE = ldap_connect(host => $self->ldap_host());
    } else {
        $LDAP_UPDATE = ldap_connect(host      => $self->ldap_host(),
                                    principal => $self->principal_name(),
                                    keytab    => $self->keytab_file(),
                                    tgt       => $self->tgt_file(),
            );
    }


    $self->log_debug($self->ldap_host()) ;
    $self->log_debug($self->principal_name()) ;
    $self->log_debug($self->keytab_file()) ;

    $self->log_debug('bind complete');
    $self->log_debug('writer bound to host: ' . $self->ldap_host()) ;

    $self->LDAP_UPDATE($LDAP_UPDATE);
    return $self->LDAP_UPDATE();
}

sub unbind_ldap_writer {
    my $self = shift;
    ldap_disconnect($self->LDAP_UPDATE());
}

sub is_private {
    my $self = shift;

    my ($visib_value) = @_;

    if (!defined($visib_value)) {
        return 0;
    } elsif (lc($visib_value) eq 'private') {
        return 1 ;
    } else {
        return 0;
    }
}


# Let S be a source array and T a target array. Consider
# S and T as "multi-sets".
#
# This function returns two arrays A and D so that
# S = T + A - D (multi-set arithmetic).
#
# Example:
#  S = (1, 1, 1, 2, 3, 5, 6, 6, 6)
#  T = (1, 1, 2, 2, 4, 6, 6, 6, 6, 6)
#
# What do you have to ADD to T and SUBTRACT to T to make it the same as S?
#
# Answer: Add: 1, 3, 5 Subtract: 2, 4, 6, 6
#
# Two hash references are returned: an "add" hash reference and a "delete"
# hash reference.
#
# From the above example, the returned hash references would be
#
#   Add:
#      1 -> 1
#      3 -> 1
#      5 -> 1
#
#   Delete:
#      2 -> 1
#      4 -> 1
#      6 -> 2

# Note that A and D are DISJOINT.
sub calc_add_delete {
    my $self = shift ;

    my ($src_aref, $trg_aref) = @_ ;

    my @A = ();
    my @D = ();

    my @S = sort @{ $src_aref };
    my @T = sort @{ $trg_aref };

    $self->log_trace('calc_add_delete');
    # The following two log commands log attribute values,
    # so we normally do not want those values written to the logs (privacy
    # reasons).We only uncomment them out in unusual circumstances
    if ($self->log_private()) {
        $self->log_trace(Dumper(\@S));
        $self->log_trace(Dumper(\@T));
    }

    my %S = ();
    my %T = ();
    hashify($src_aref, \%S);
    hashify($trg_aref, \%T);

    my @all_keys = (keys %S, keys %T);

    my %unique_keys = () ;
    foreach my $ak (@all_keys) {
        $unique_keys{$ak} = 1;
    }

    my %A = ();
    my %D = ();

    foreach my $key (keys %unique_keys) {
        if (!exists($T{$key})) {
            $T{$key} = 0;
        }
        if (!exists($S{$key})) {
            $S{$key} = 0;
        }

        if ($S{$key} > $T{$key}) {
            $A{$key} = $S{$key} - $T{$key};
        } elsif ($S{$key} < $T{$key}) {
            $D{$key} = $T{$key} - $S{$key};
        } else {
            # They are the same, so nothing to do.
        }
    }

    return (\%A, \%D) ;
}

sub hashify {
    my ($array_ref, $hash_ref) = @_ ;

    my @vals = sort @{ $array_ref };

    my $counter = 0;
    my $cur_val = undef;
    foreach my $v (@vals) {
        if (defined($cur_val) && ($cur_val ne $v)) {
            # New value
            $hash_ref->{$cur_val} = $counter ;
            $counter = 1;
        } else {
            # Same value
            ++$counter;
        }
        $cur_val = $v;
    }
    if (defined($cur_val)) {
        $hash_ref->{$cur_val} = $counter;
    }

    return ;
}


# sync all entries manually
sub sync_entries {
    my $self = shift;
    my ($dns_to_sync_aref) = @_ ;

    $self->log_debug('calling: sync_entries');

    my @dns = @{ $dns_to_sync_aref } ;

    foreach my $dn (@dns) {
        my $updated = $self->sync_entry($dn);
        $self->pause();
    }

    $self->log_debug('exiting: sync_entries');

    return;
}

# Search the directory and get the person dn. This has the form
#
#      suRegID=111823c0c3fe409e9fbde77fee52b100,cn=people,dc=stanford,dc=edu
#
# Note that if we search the people tree we simply grab "dn". If we
# search the accounts tree we grab "owner".
sub get_all_entries_to_sync {
    my $self = shift;

    $self->log_indent_inc();
    $self->log_debug('calling: get_all_entries_to_sync');

    my $base_tree_filter = $self->base_tree_filter();
    my $scope            = LDAP_SCOPE_SUBTREE;
    my $filter           = q{(} . $self->in_filter() . q{)};

    my @attrs ;
    my $base ;
    my $is_accounts ;
    if ($base_tree_filter =~ m{accounts}ixsm) {
        # We are searching against the accounts tree, so the attribute we
        # want is "owner"
        $base        = 'cn=accounts,dc=stanford,dc=edu' ;
        $is_accounts = 1 ;
        @attrs       = ('owner');
        $self->log_debug('base tree is accounts');
    } elsif ($base_tree_filter =~ m{people}ixsm) {
        # We are searching against the people tree, so the attribute we
        # want is "objectclass" (dn comes free).
        $base = 'cn=people,dc=stanford,dc=edu' ;
        $is_accounts = 0 ;
        @attrs  = ('objectclass');
        $self->log_debug('base tree is people');
    } else {
        croak q{base_tree_filter must be one of 'accounts' or 'people'} ;
    }

    my $LDAP = $self->LDAP();

    # This array will contain the DN's of the entries from the people tree
    # we want to sync.
    my @entry_dns = () ;

    $self->log_debug('sync_entries starting ldap search');
    $self->log_debug("parameters base: $base, scope: $scope, filter: $filter");
    $self->log_debug('please stand by...');
    my $stat = $LDAP->search_s($base, $scope, $filter, \@attrs, 0);
    $self->log_debug('sync_entries finished ldap search');
    if ($stat == LDAP_NO_SUCH_OBJECT) {
        $self->log_warn("No entries found. Base:$base Filter:$filter Scope:$scope");
    } elsif ($stat != LDAP_SUCCESS) {
        my $m ;
        if ($LDAP->errstring) {
            $m .= $LDAP->errstring;
        }
        if ($LDAP->extramsg) {
            $m .= ' - '.$LDAP->extramsg;
        }
        if (!$m) {
            $m = "Unknown LDAP error status = $stat";
        }
        $self->log_fatal($m);
    } else {
        my $entry_ref = $LDAP->get_all_entries;
        my %entry = %{ $entry_ref };
        if ($is_accounts) {
            # filtering on the ACCOUNTS tree

            # Extract all the "owner" attributes.
            my @entries = keys %entry;
            foreach my $entry_key (@entries) {
                my $value_href = $entry{$entry_key} ;

                my $owners_aref ;
                if (exists($value_href->{'owner'})) {
                    $owners_aref = $value_href->{'owner'} ;
                } else {
                    $self->log_warn("skipping entry $entry_key: no owner attribute");
                    next ;
                }

                my @owners = @{ $owners_aref  };
                my $owner ;
                if (scalar(@owners) == 0) {
                    $self->log_warn("skipping entry $entry_key: zero owner attributes");
                    next ;
                } elsif (scalar(@owners) > 1) {
                    $self->log_warn("skipping entry $entry_key: more than one owner attribute");
                    next ;
                } else {
                    $owner = $owners[0] ;
                }

                # We only want to act on accounts owned by people, so if
                # the owner of this account is not a person, skip
                if ($owner !~ m{cn=people,dc=stanford,dc=edu}xsm) {
                    $self->log_debug("skipping entry $entry_key: the owner ($owner) is not in the people tree");
                } else {
                    push(@entry_dns, $owner) ;
                }
            }
        } else {
            # filtering on the PEOPLE tree
            @entry_dns = keys %entry;
            undef(%entry);
        }

        my $number_entries = scalar(@entry_dns);
        $self->log_debug("sync_entries found $number_entries entries");
    }

    $self->log_debug('exiting: get_all_entries_to_sync');
    $self->log_indent_dec();

    return @entry_dns ;
}


# Given a user's dn, return an array of that user's relevant
# attributes. The values returned are the values of the attributes
# mentioned in $self->VISIB_ATTRIBUTES.
#
# If the user cannot be found, returns the empty array.
#
# This function dies on any other error.
sub get_all_entries_from_dn {
    my $self = shift;
    $self->log_indent_inc();

    my ($dn) = @_ ;

    my %SRCATTR_TO_TARGATTR = %{ $self->SRCATTR_TO_TARGATTR() };
    my %VISIB_ATTRIBUTES    = %{ $self->VISIB_ATTRIBUTES() };
    my $LDAP                = $self->LDAP();

    my @source_attrs = keys   %SRCATTR_TO_TARGATTR;
    my @target_attrs = values %SRCATTR_TO_TARGATTR;
    my @visib_attrs  = keys   %VISIB_ATTRIBUTES;
    my @attrs        = (@source_attrs, @target_attrs, @visib_attrs);
    $self->log_trace('source attributes: ' . join(q{,}, @source_attrs));
    $self->log_trace('target attributes: ' . join(q{,}, @target_attrs));
    $self->log_trace('visib  attributes: ' . join(q{,}, @visib_attrs));
    $self->log_trace('all attributes: '    . join(q{,}, @attrs));

    # Get the entry corresponding to $dn.
    my $in_scope = LDAP_SCOPE_BASE;
    $self->log_trace("\$in_scope: $in_scope");

    my $ldap_filter = q{(objectClass=*)};
    $self->log_trace("\$ldap_filter: $ldap_filter");

    $self->log_debug("search parameters: base:$dn filter:$ldap_filter scope:$in_scope");
    my $stat = $LDAP->search_s($dn, $in_scope, $ldap_filter, \@attrs, 0);
    $self->log_trace('search completed');

    my $in_cnt;
    my %entry;
    if ($stat == LDAP_NO_SUCH_OBJECT) {
        $self->log_debug("sync dn '$dn' entry not found");
        %entry = ();
        return %entry;
    } elsif ($stat != LDAP_SUCCESS) {
        $self->log_warn("LDAP search error ($stat) "
              . "Base:$dn Filter:$ldap_filter Scope:$in_scope");
        my $m;
        if ($LDAP->errstring) {
            $m .= $LDAP->errstring;
        }
        if ($LDAP->extramsg) {
            $m .= ' - '.$LDAP->extramsg;
        }
        if (!$m) {
            $m = "Unknown LDAP error status = $stat";
        }
        die $m;
    } else {
        $self->log_debug('input entry found');
        my $entry_ref = $LDAP->get_all_entries;
        $self->log_trace('extracted all entries');

        %entry = %{ $entry_ref };
        $in_cnt = scalar(keys %entry);
        $self->log_debug("number of keys in entry: $in_cnt");
    }

    # Issue warnings if we get anything other than one entry back and
    # bail out.
    if ($in_cnt > 1) {
        $self->log_warn("Duplicates found for $dn ... skipping" . Dumper(%entry));
        $self->log_indent_dec();
        %entry = ();
        return %entry;
    } elsif ($in_cnt == 0) {
        $self->log_debug("Entry not found for $dn");
        $self->log_indent_dec();
        %entry = ();
        return %entry;
    }

    # If we get here we know that %entry has only one entry.
    # Get the input dn of this one entry.
    my $entry_dn = (keys %entry)[0] ;
    $self->log_debug("found dn: $entry_dn");

    # Get the attributes;
    my %entry_attributes = %{$entry{$entry_dn}} ;

    $self->log_indent_dec();
    return %entry_attributes;
}

# Given a single dn, sync the suGALperson attributes from the person
# attributes.
#
# Returns -1 on failure.
sub sync_entry {
    my $self = shift;
    $self->log_indent_inc();

    $self->log_trace('calling: sync_entry');

    my ($sync_dn) = @_;
    $self->log_debug("syncing $sync_dn");

    # Get all the attribute values for $sync_dn.
    my %entry_attributes = $self->get_all_entries_from_dn($sync_dn);

    # Get an identifier
    my $identifier = $sync_dn;
    if (exists($entry_attributes{'uid'})) {
        if ($entry_attributes{'uid'}->[0]) {
            $identifier = $entry_attributes{'uid'}->[0];
        }
    } elsif (exists($entry_attributes{'cn'})) {
        if ($entry_attributes{'cn'}->[0]) {
            $identifier = $entry_attributes{'cn'}->[0];
        }
    }

    ##################################################################
    ##################################################################

    # Now we create/delete/update all the target attributes with the
    # source attributes taking into account the visibility settings.
    my %ldap_adds          = ();
    my %ldap_deletes       = ();
    my %ldap_objectclasses = ();

    my %VISIB_ATTRIBUTES    = %{ $self->VISIB_ATTRIBUTES() };

    foreach my $visib_attr (keys %VISIB_ATTRIBUTES) {
        $self->log_debug("processing attributes controlled by $visib_attr");
        my $attrs_to_gal_attrs_href = $VISIB_ATTRIBUTES{$visib_attr};

        my $visib_value = $self->get_single_attribute_from_hash(\%entry_attributes, $visib_attr);
        if ($visib_value) {
            $self->log_debug("$visib_attr has value $visib_value");
        } else {
            $self->log_debug("$visib_attr attribute does not exist");
        }

        if ($self->is_private($visib_value)) {
            $self->log_debug("visibility attribute '$visib_attr' is private");
            # This visibility attribute is set to 'private', so make sure all the
            # suGAL attributes are undefined.
            my @sugalattrs = values %{ $attrs_to_gal_attrs_href };
            foreach my $su_gal_attr (@sugalattrs) {
                $self->log_trace("verifying that suGAL attribute '$su_gal_attr' is not defined");
                my @su_gal_values = $self->get_multiple_attributes_from_hash(\%entry_attributes, $su_gal_attr);
                if (@su_gal_values) {
                    # There are some values here.
                    if ($self->log_private) {
                        $self->log_debug("adding $su_gal_attr with some value(s) '"
                                        . join(q{,}, @su_gal_values) . q{' to DELETE queue});
                    } else {
                        my $number_to_delete = scalar(@su_gal_values);
                        $self->log_debug("deleting from $su_gal_attr $number_to_delete value(s)");
                    }
                    # If we put in "undef" for the values, this should
                    # delete all instances of $su_gal_attr.
                    push(@{$ldap_deletes{$su_gal_attr}{'d'}}, undef);
                } else {
                    $self->log_trace("no values found for suGAL attribute '$su_gal_attr'");
                }
            }
        } else {
            $self->log_debug("visibility attribute '$visib_attr' is NOT private");
            # This visibility attribute is NOT set to 'private', so make sure all the
            # suGAL attributes are synchronized.
            my @attrs1 = keys %{ $attrs_to_gal_attrs_href };
            foreach my $attr (@attrs1) {
                my $su_gal_attr = $attrs_to_gal_attrs_href->{$attr};
                $self->log_debug("checking to see if suGAL attribute '$su_gal_attr' is already properly set");

                my @values        = $self->get_multiple_attributes_from_hash(\%entry_attributes, $attr);
                my @su_gal_values = $self->get_multiple_attributes_from_hash(\%entry_attributes, $su_gal_attr);

                my ($add_href, $delete_href) = $self->calc_add_delete(\@values, \@su_gal_values);

                if ($self->log_private) {
                    $self->log_trace("values to ADD to $su_gal_attr: " . Dumper($add_href));
                    $self->log_trace("values to DELETE to $su_gal_attr: " . Dumper($delete_href));
                } else {
                    my $number_adding   = scalar(keys %{ $add_href });
                    my $number_deleting = scalar(keys %{ $delete_href });
                    $self->log_trace("values to ADD to $su_gal_attr: $number_adding value(s)");
                    $self->log_trace("values to DELETE to $su_gal_attr: $number_deleting value(s)");
                }

                # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### #
                # Add the action to either the 'add' or 'deletes' queue.
                my $add_action_fref = sub {
                    my ($action_href, $action) = @_ ;

                    my ($ldap_add_code, $action_queue, $ldap_queue_href);
                    if ($action eq 'add') {
                        $ldap_add_code   = 'a';
                        $action_queue    = 'ADD';
                        $ldap_queue_href = \%ldap_adds;
                    } elsif ($action eq 'delete') {
                        $ldap_add_code   = 'd';
                        $action_queue    = 'DELETE';
                        $ldap_queue_href = \%ldap_deletes;
                    } else {
                        croak "unknown action '$action'";
                    }

                    foreach my $value (keys %{ $action_href }) {
                        # How many to process?
                        my $number_to_process = $action_href->{$value};
                        my @add_values = ();
                        for my $i (1..$number_to_process) {
                            push(@add_values, $value);
                        }
                        if ($self->log_private()) {
                            $self->log_debug("adding $su_gal_attr with value(s) '"
                                            . join(q{,}, @add_values) . qq{' to $action_queue queue});
                        } else {
                            $self->log_debug("adding to $su_gal_attr $number_to_process value(s)");
                        }
                        push(@{$ldap_queue_href->{$su_gal_attr}{$ldap_add_code}}, @add_values);
                    }
                } ;
                # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### # # ### #

                # Add attributes to the delete and/or add queue.
                $self->log_debug("attribute is [$attr]");
                if ($attr =~ m{^facsimileTelephoneNumber$}xsm) {
                    $self->log_debug('syncing special attribute facsimileTelephoneNumber');
                    # We have to do something special for
                    # suGALfacsimileTelephoneNumber. This attribute requires
                    # all the attributes to FIRST be deleted, and THEN add
                    # back the ones ones you want to keep.

                    # If we either need to delete or add, let's delete
                    # everything and then add back.
                    if ((%{ $add_href }) || (%{ $delete_href })) {
                        $self->log_debug('add_href: ' . Dumper($add_href));
                        $self->log_debug('delete_href: ' . Dumper($delete_href));

                        # If there are any suGAL values, delete them.
                        if (@su_gal_values) {
                            push(@{$ldap_deletes{$su_gal_attr}{'d'}}, undef);
                        }

                        # If there are any values, add them
                        if (@values) {
                            push(@{$ldap_adds{$su_gal_attr}{'a'}}, @values);
                        }
                    }
                } else {
                    # Everything else.
                    if (%{ $add_href }) {
                        $add_action_fref->($add_href, 'add');
                    }

                    if (%{ $delete_href }) {
                        $add_action_fref->($delete_href, 'delete');
                    }
                }

            }
        }
    }

    if ($self->log_private()) {
        $self->log_debug(%ldap_adds);
    }

    # If the number of target attributes to add is greater than zero, make
    # sure that we add an suGALperson objectClass, if the objectClass is
    # not already present.
    my $add_object_class_flag = 0;
    if (scalar(keys %ldap_adds)) {
        my $cnt = $self->get_entry_cnt($sync_dn, 'objectClass=suGALperson', LDAP_SCOPE_BASE);
        if (!$cnt) {
            push @{$ldap_objectclasses{'objectClass'}{'a'}}, 'suGALperson';
            $add_object_class_flag = 1;
            if ($self->update) {
                $self->log_info("($identifier): will add objectClass suGALperson");
            } else {
                $self->log_debug("($identifier): if running in update mode would have added objectClass suGALperson");
            }
        }
    }

    # Update the attributes.
    my $update_flag = 0;
    if ($add_object_class_flag) {
        $self->log_debug("($identifier): adding object class...");
        $self->update_ldap($sync_dn, $identifier, \%ldap_objectclasses);
        ++$update_flag;
    }

    if (scalar(keys %ldap_deletes) > 0) {
        $self->log_debug("($identifier): deleting...");
        $self->update_ldap($sync_dn, $identifier, \%ldap_deletes);
        ++$update_flag;
    }

    if (scalar(keys %ldap_adds) > 0) {
        $self->log_debug("($identifier): adding...");
        $self->update_ldap($sync_dn, $identifier, \%ldap_adds);
        ++$update_flag;
    }

    if (!$update_flag) {
        $self->log_debug("no changes needed for entry ($identifier)($sync_dn)");
    } elsif ($self->update) {
        $self->log_info("($identifier): person attributes synchronized to suGAL") ;
    } else {
        $self->log_debug("($identifier): if running in update mode, person attributes would have synchronized to suGAL") ;
    }


    $self->log_trace('exiting sync_entry');

    $self->log_indent_dec();
    return $update_flag;
}

sub update_ldap {
    my $self = shift;

    $self->log_indent_inc();

    $self->log_trace('entering update_ldap');

    my ($dn, $identifier, $mods_href) = @_;

    my %mods = %{ $mods_href };

    my $LDAP_UPDATE = $self->LDAP_UPDATE();

    if (!$LDAP_UPDATE) {
        $self->bind_ldap_writer();
        $LDAP_UPDATE = $self->LDAP_UPDATE();
    }

    $self->log_debug("update_ldap: dn is $dn");

    if (keys %mods) {
        $self->log_trace('there are modifications to make');
        if ($self->update()) {
            $self->log_trace('the update flag is set');
            my $stat = $LDAP_UPDATE->modify_s($dn, \%mods);
            if ($stat != LDAP_SUCCESS) {
                $self->log_warn("LDAP update error dn:$dn");
                my $m = $LDAP_UPDATE->errstring;
                if ($LDAP_UPDATE->extramsg) {
                    $m .= ' - '.$LDAP_UPDATE->extramsg;
                }
                $m .= Dumper(%mods);
                $self->log_error($m);
            } else {
                $self->log_debug("$identifier: synchronized suGAL attributes");
            }
        } else {
            print "  Proposing update to $dn\n";
            foreach my $attr (keys %mods) {
                my $action_href = $mods{$attr};
                my $action = (keys %{ $action_href })[0];

                my @values = map { q{'} . $_ . q{'} } @{$action_href->{$action}};

                my $value = join(q{ and }, @values) ;

                my $msg ;
                if ($action eq 'a') {
                    $msg = "will add attribute $attr with value(s) $value";
                } elsif ($action eq 'r') {
                    $msg = "will replace attribute $attr with value(s) $value";
                } elsif ($action eq 'd') {
                    $msg = "will delete attribute $attr with value(s) $value";
                } else {
                    $msg = "unknown action '$action'";
                    $self->log_error($msg);
                }
                print "    $msg\n";
                $self->log_debug("update_ldap (dry run): $msg");
            }
        }
    } else {
        $self->log_trace("$identifier: there are NO modifications to make");
    }

    $self->log_trace('leaving update_ldap');

    $self->log_indent_dec();

    return;
}


sub listen_for_changes {
    my $self = shift;

    $self->is_listener(1);

    $self->log_indent_inc();

    # Construct the list of attributes that we listen to for changes.

    # We want to listen to the person object attributes NOT the
    # suGALPerson object attributes. We also need to listen to the
    # visibility-control attributes.
    my %SRCATTR_TO_TARGATTR = %{ $self->SRCATTR_TO_TARGATTR() };
    my @source_attrs = keys %SRCATTR_TO_TARGATTR ;

    # Get the visibility-control attributes.
    my %VISIB_ATTRIBUTES = %{ $self->VISIB_ATTRIBUTES() };
    my @visibility_attributes = keys %VISIB_ATTRIBUTES;
    @source_attrs = (@source_attrs, @visibility_attributes);

    my %source_attrs_indicator = () ;
    foreach my $source_attr (@source_attrs) {
        $source_attrs_indicator{$source_attr} = 1;
    }
    $self->log_trace('source attributes to listen for: ' . join(q{,}, @source_attrs));

    my $tree_dn = uc(Stanford::LDAPserver::squeeze($self->people_dn));

    my $log_base_dn     = $self->log_base_dn();
    my $log_filter      = $self->log_filter();
    my $log_cookie_file = $self->log_cookie_file();

    $self->log_debug("listener parameters: base: $log_base_dn, filter: $log_filter, cookie: $log_cookie_file");

    # If the cookie file does not exist, create a new one.
    if (!(-e $log_cookie_file)) {
        $self->log_debug('no cookie file, so create one');
        open(my $FH, q{>}, $log_cookie_file);
        close($FH);
    }

    my $LDAP = $self->LDAP();

    $self->log_debug('create the listener');
    # WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!
    # Whatever you do, do NOT, I repeat, NOT put the listen_for_changes
    # inside an eval{}. (There's a weekend I am never getting back.)
    # WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!
    my $msgid = $LDAP->listen_for_changes(-basedn  => $log_base_dn,
                                          -scope   => LDAP_SCOPE_SUBTREE,
                                          -filter  => $log_filter,
                                          -cookie  => $log_cookie_file);

    $self->log_debug("listener created; msgid: $msgid");
    my %retry_list = ();

    # Loop forever, listening for changes.  Every time one is found,
    # iterate through each result, and make changes as needed.

    while (1) {
        $self->log_trace('at start of listen loop');
        my @entries;
        while (!(-e '/etc/sync-sugal-stop') && (@entries = $LDAP->next_changed_entries($msgid, 0, -1))) {
            $self->log_debug('listener has detected a change');
            foreach my $entry (@entries) {

                # Find the dn of whatever changed
                my @reqdn = $LDAP->get_values('reqDN');
                if (!@reqdn) {
                    $self->log_debug('reqdn came back empty so skipping');
                    next;
                }
                my $dn = $reqdn[0];
                $self->log_debug("listener changed entry has dn '$dn'");

                # Compare DNs to see if it is one that we are handling
                my $test_dn = uc(Stanford::LDAPserver::squeeze($dn));
                if ($test_dn !~ m{$tree_dn$}ixsm) {
                    $self->log_debug("the dn '$dn' is not in our tree ('$test_dn'), so skip");
                    next;
                }

                $self->log_debug("processing $dn");

                # Look at the changes
                my @mods = ();
                my $update_request = 0;
                if (@mods = $LDAP->get_values('reqMod')) {
                    foreach my $val (@mods) {
                        if ($self->log_private()) {
                            $self->log_debug("looking at a change value '$val'");
                        } else {
                            $self->log_debug('looking at a change value (hidden for privacy reason)');
                        }
                        # Pick apart the result
                        my $attr = my $action = my $value = q{};
                        if ($val =~ m{^(.+?):(.)\s+(.+)$}xsm) {
                            $attr   = $1;
                            $action = $2;
                            $value  = $3;
                            if ($self->log_private()) {
                                $self->log_debug("attribute: [$attr], action: [$action], value: [$value]");
                            } else {
                                $self->log_debug("attribute: [$attr], action: [$action], value: [hidden]");
                            }
                        } elsif ($val =~ m{^(.+?):-}xsm) {
                            $attr = $1;
                            $self->log_debug("attribute: [$attr], action: [delete], value: NONE");
                        } elsif ($val =~ m{^([^:]+):=$}xsm) {
                            $attr = $1;
                            $self->log_debug("attribute: [$attr], action: set to empty");
                        }

                        if (!$attr) {
                            $self->log_debug('changed attribute empty, so skipping');
                            next;
                        }

                        # Check to see if the attribute modified is
                        # one that we want replicated.
                        if ($source_attrs_indicator{$attr}) {
                            if ($attr =~ m{suVisib}xsm) {
                                $self->log_info("attribute '$attr' has changed and the attributes it controls need to be replicated");
                            } else {
                                $self->log_info("attribute '$attr' has changed and needs to be replicated");
                            }
                            $update_request = 1;
                        }
                    }
                }

                if ($update_request) {
                    # Try to sync entries.  If that fails then put it on
                    # the retry list to try again.
                    if ($self->sync_entry($dn) < 0) {
                        $self->log_warn("sync_entry failed for $dn; putting on retry list");
                        if ($retry_list{$dn}) {
                            $retry_list{$dn}++;
                        } else {
                            $retry_list{$dn} = 1;
                        }
                    }
                    foreach my $retry_dn (keys %retry_list) {
                        if ($self->sync_entry($retry_dn)) {
                            delete ($retry_list{$retry_dn});
                        } elsif ($retry_list{$retry_dn}++ > $self->retry_count()) {
                            delete($retry_list{$retry_dn});
                            $self->log_warn("No updates performed for $dn");
                        }
                    }
                }
            }

            # Disconnect when the updates are complete
            $self->log_trace('disconnect ldap update');
            my $LDAP_UPDATE = $self->LDAP_UPDATE();
            if ($LDAP_UPDATE) {
                ldap_disconnect($LDAP_UPDATE);
                undef $LDAP_UPDATE;
                $self->LDAP_UPDATE(undef);
            }

            $self->log_trace('freeing message');
            $LDAP->msgfree;
        }

    }

    $self->log_info('exited listener loop');
    $self->log_indent_dec();
    return;
}


# Returns:
#  * the attribute value, if there is exactly one attribute with a value;
#  * undef, if there is not attribute with the supplied name;
#  * dies, if there are more than one attributes with the supplied name.
#
# Needs an existing entry attributes hash.
sub get_single_attribute_from_hash {
    my $self = shift;

    my ($attrs_href, $attribute_name) = @_;

    my $attr_values_aref = $attrs_href->{$attribute_name};

    if (!defined($attr_values_aref)) {
        return undef ; ## no critic (ProhibitExplicitReturnUndef)
    }

    my @values = @{ $attr_values_aref };
    if (scalar(@values) > 1) {
        my $m = "single-valued attribute $attribute_name has more than value";
        $self->log_fatal($m);
    } elsif (scalar(@values) == 0) {
        return undef ; ## no critic (ProhibitExplicitReturnUndef)
    } else {
        my $value = $values[0];
        return $value ;
    }
    return;
}

sub get_multiple_attributes_from_hash {
    my $self = shift;

    my ($attrs_href, $attribute_name) = @_;

    my $attr_values_aref = $attrs_href->{$attribute_name};

    if (defined($attr_values_aref)) {
        return @{ $attr_values_aref };
    } else {
        return () ;
    }
    return;
}

## UTILITY (for testing code)

# These read and write using LDAP_UPDATE connection.
sub set_attribute {
    my $self = shift;

    my ($dn, $attribute, $values_aref) = @_ ;
    my %ldap_mods = () ;


    # ### #    # ### #    # ### #    # ### #    # ### #    # ### #    #
    my $parse_status_fref = sub {
        my ($status) = @_ ;

        if ($status != LDAP_SUCCESS) {
            my $m = $self->LDAP_UPDATE->errstring;
            if ($self->LDAP_UPDATE->extramsg) {
                $m .= ' - '.$self->LDAP_UPDATE->extramsg;
            }
            warn $m;
            return 0;
        } else {
            return 1;
        }
    } ;
    # ### #    # ### #    # ### #    # ### #    # ### #    # ### #    #


    # If the attribute is facsimileTelephoneNumber or
    # suGALfacsimileTelephoneNumber, we first DELETE and then ADD. We have
    # to do this because these attributes have no EQUALITY matching rules.
    if ($attribute =~ /^(facsimileTelephoneNumber|suGALfacsimileTelephoneNumber)$/ixsm) {
        my @values = $self->get_attribute($dn, $attribute);

        my $stat;

        # delete, but only if there are some values.
        if (@values) {
            push(@{$ldap_mods{$attribute}{'d'}}, undef);
            $stat = $self->LDAP_UPDATE()->modify_s($dn, \%ldap_mods);
            if (!($parse_status_fref->($stat))) {
                return 0;
            }
        } else {
            # print "no need to delete since no $attribute values\n";
        }


        # add
        %ldap_mods = () ; # Need to zero this out.
        push(@{$ldap_mods{$attribute}{'a'}}, @{ $values_aref });
        $stat = $self->LDAP_UPDATE()->modify_s($dn, \%ldap_mods);
        if (!($parse_status_fref->($stat))) {
            return 0;
        }

    } else {

        push(@{$ldap_mods{$attribute}{'r'}}, @{ $values_aref });

        my $stat = $self->LDAP_UPDATE()->modify_s($dn, \%ldap_mods);
        return $parse_status_fref->($stat);
    }

    return 1;
}
sub add_attribute {
    my $self = shift;

    my ($dn, $attribute, $values_aref) = @_ ;
    my %ldap_mods = () ;

    push(@{$ldap_mods{$attribute}{'a'}}, @{ $values_aref });

    my $status = $self->LDAP_UPDATE()->modify_s($dn, \%ldap_mods);

    if ($status != LDAP_SUCCESS) {
        my $m = $self->LDAP_UPDATE->errstring;
        if ($self->LDAP_UPDATE->extramsg) {
            $m .= ' - '.$self->LDAP_UPDATE->extramsg;
        }
        warn $m;
        return 0;
    } else {
        return 1;
    }

}

sub delete_attribute {
    my $self = shift;

    my ($dn, $attribute, $values_aref) = @_ ;
    my %ldap_mods = () ;
    push(@{$ldap_mods{$attribute}{'d'}}, @{ $values_aref });

    my $stat = $self->LDAP_UPDATE()->modify_s($dn, \%ldap_mods);
    if ($stat != LDAP_SUCCESS) {
        return 0;
    } else {
        return 1;
    }
}

sub add_object_class {
    my $self = shift;
    my ($dn, $object_class) = @_;

    my $cnt = $self->get_entry_cnt($dn, "objectClass=$object_class", LDAP_SCOPE_BASE);
    if (!$cnt) {
        my %ldap_objectclasses = ();
        push(@{$ldap_objectclasses{'objectClass'}{'a'}}, $object_class);
        $self->update_ldap($dn, \%ldap_objectclasses);
        return 1;
    } else {
        return 0;
    }
}

# Returns an array.
sub get_attribute {
    ## no critic (Capitalization)
    my $self = shift;

    my ($dn, $attribute) = @_ ;

    my $suRegID = $dn;
    $suRegID =~ s{^suRegID=([\da-f]+),.*$}{$1}xsm;

    my $base   = $self->people_dn();
    my $scope  = LDAP_SCOPE_SUBTREE;
    my $filter = "suRegID=$suRegID";
    my @attrs  = ( $attribute );

    my $stat      = $self->LDAP_UPDATE()->search_s($base, $scope, $filter, \@attrs,0);
    my $entry_ref = $self->LDAP_UPDATE()->get_all_entries;
    if ($stat != LDAP_SUCCESS) {
        return ();
    } else {
        if (wantarray()) {
            my $values_aref = ($entry_ref->{$dn})->{$attribute};
            if (defined($values_aref)) {
                return @{ $values_aref };
            } else {
                return ();
            }
        } else {
            my $value = ($entry_ref->{$dn})->{$attribute}->[0];
            return $value;
        }
    }

}


__PACKAGE__->meta->make_immutable;

1;

__END__


=head1 NAME

Stanford::LDAP::SyncSuGAL - Sync attributs to suGALPerson

=head1 VERSION

Version 0.01

=head1 ABSTRACT

Sync attributs to suGALPerson ...

=head1 CONFIGURATION

The configuration file is a simple attribute=value format text file.

This example contains all of the valid properties that can be
specified and shows the default settings for each property.

    # Host to update
    ldap_host = ldap.stanford.edu
    #
    # Where to find the people data
    people_dn = cn=people,dc=stanford,dc=edu
    #
    # Filter: specify the base tree and a filter.
    # base_tree_filter should be set to either 'accounts' or 'people'.
    base_tree_filter = accounts
    in_filter        = suEmailStatus=active
    #
    #
    # Listen parameters for reading the ldap transaction log (cn=accesslog)
    log_base_dn = cn=accesslog
    log_filter  = objectclass=*
    log_cookie  = /var/run/ldap-sync-attribute.cookie
    #
    # A file that lists the mapping of visibility controller attributes to
    # the attributes they control. Normally, this setting should not be
    # changed.
    visibility_attributes_file = /etc/ldapadmin/sugal-visibility.yaml
    #
    #
    # Kerberos parameters.  There is no default for keytab_file or
    # principal_name.  If a keytab file is specified then a principal
    # must be specified and a ticket cache is created.  It the tgt_file
    # is not specified, then /tmp/krb5_<scriptname>_<pid>.tgt will be used.
    keytab_file = /etc/file.keytab
    principal_name = service/some_service
    # tgt_file =
    #
    # If you want to ignore keytab and TGT files and depend instead on
    # the KRB5CCNAME environment variable, set use_krb5ccname to 1.
    # use_krb5ccname = 1
    #
    # Specify the log4perl configuration file; if not specified, a default
    # configuration will be used that

    # If an LDAP modify fails, how many times to retry. The default value
    # is 5.
    retry_count = 2


=head1 AUTHOR

Adam Lewenberg, C<< <adamhl at stanford.edu> >>

=head1 BUGS

Please report any bugs or feature requests
to C<adamhl at stanford.edu>.

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Stanford::LDAP::SyncSuGAL

=head1 COPYRIGHT & LICENSE

Copyright 2015 Stanford University, all rights reserved.

=cut
